"""Support for Netatmo energy devices (relays, thermostats and valves)."""
import logging
from abc import ABC
from collections import defaultdict
from typing import Any, Dict, Optional

from .auth import AbstractAsyncAuth, NetatmoOAuth2
from .exceptions import InvalidRoom, NoDevice, NoSchedule
from .helpers import _BASE_URL

LOG = logging.getLogger(__name__)

_GETHOMESDATA_REQ = _BASE_URL + "api/homesdata"
_GETHOMESTATUS_REQ = _BASE_URL + "api/homestatus"
_SETTHERMMODE_REQ = _BASE_URL + "api/setthermmode"
_SETROOMTHERMPOINT_REQ = _BASE_URL + "api/setroomthermpoint"
_GETROOMMEASURE_REQ = _BASE_URL + "api/getroommeasure"
_SWITCHHOMESCHEDULE_REQ = _BASE_URL + "api/switchhomeschedule"


class AbstractHomeData(ABC):
    """Abstract class of Netatmo energy devices."""

    raw_data: Dict = defaultdict(dict)
    homes: Dict = defaultdict(dict)
    modules: Dict = defaultdict(dict)
    rooms: Dict = defaultdict(dict)
    schedules: Dict = defaultdict(dict)
    zones: Dict = defaultdict(dict)
    setpoint_duration: Dict = defaultdict(dict)

    def process(self) -> None:
        """Process data from API."""
        self.homes = {d["id"]: d for d in self.raw_data}

        for item in self.raw_data:
            home_id = item.get("id")
            home_name = item.get("name")

            if not home_name:
                home_name = "Unknown"
                self.homes[home_id]["name"] = home_name

            if "modules" not in item:
                continue

            for module in item["modules"]:
                self.modules[home_id][module["id"]] = module

            self.setpoint_duration[home_id] = item.get(
                "therm_setpoint_default_duration",
            )

            for room in item.get("rooms", []):
                self.rooms[home_id][room["id"]] = room

            for schedule in item.get("schedules", []):
                schedule_id = schedule["id"]
                self.schedules[home_id][schedule_id] = schedule

                if schedule_id not in self.zones[home_id]:
                    self.zones[home_id][schedule_id] = {}

                for zone in schedule["zones"]:
                    self.zones[home_id][schedule_id][zone["id"]] = zone

    def _get_selected_schedule(self, home_id: str) -> Dict:
        """Get the selected schedule for a given home ID."""
        for value in self.schedules.get(home_id, {}).values():
            if "selected" in value.keys():
                return value

        return {}

    def get_hg_temp(self, home_id: str) -> Optional[float]:
        """Return frost guard temperature value."""
        return self._get_selected_schedule(home_id).get("hg_temp")

    def get_away_temp(self, home_id: str) -> Optional[float]:
        """Return the configured away temperature value."""
        return self._get_selected_schedule(home_id).get("away_temp")

    def get_thermostat_type(self, home_id: str, room_id: str) -> Optional[str]:
        """Return the thermostat type of the room."""
        for module in self.modules.get(home_id, {}).values():
            if module.get("room_id") == room_id:
                return module.get("type")

        return None

    def is_valid_schedule(self, home_id: str, schedule_id: str):
        """Check if valid schedule."""
        schedules = {
            self.schedules[home_id][s]["name"]: self.schedules[home_id][s]["id"]
            for s in self.schedules.get(home_id, {})
        }
        return schedule_id in list(schedules.values())


class HomeData(AbstractHomeData):
    """Class of Netatmo energy devices."""

    def __init__(self, auth: NetatmoOAuth2) -> None:
        """Initialize the Netatmo home data.

        Arguments:
            auth {NetatmoOAuth2} -- Authentication information with a valid access token
        """
        self.auth = auth

    def update(self) -> None:
        """Fetch and process data from API."""
        resp = self.auth.post_request(url=_GETHOMESDATA_REQ)
        if resp is None or "body" not in resp:
            raise NoDevice("No thermostat data returned by Netatmo server")

        self.raw_data = resp["body"].get("homes")
        if not self.raw_data:
            raise NoDevice("No thermostat data available")

        self.process()

    def switch_home_schedule(self, home_id: str, schedule_id: str) -> Any:
        """Switch the schedule for a give home ID."""
        if not self.is_valid_schedule(home_id, schedule_id):
            raise NoSchedule("%s is not a valid schedule id" % schedule_id)

        post_params = {"home_id": home_id, "schedule_id": schedule_id}
        resp = self.auth.post_request(url=_SWITCHHOMESCHEDULE_REQ, params=post_params)
        LOG.debug("Response: %s", resp)


class AsyncHomeData(AbstractHomeData):
    """Class of Netatmo energy devices."""

    def __init__(self, auth: AbstractAsyncAuth) -> None:
        """Initialize the Netatmo home data.

        Arguments:
            auth {AbstractAsyncAuth} -- Authentication information with a valid access token
        """
        self.auth = auth

    async def async_update(self):
        """Fetch and process data from API."""
        resp = await self.auth.async_post_request(url=_GETHOMESDATA_REQ)
        if resp is None or "body" not in resp:
            raise NoDevice("No thermostat data returned by Netatmo server")

        self.raw_data = resp["body"].get("homes")
        if not self.raw_data:
            raise NoDevice("No thermostat data available")

        self.process()

    async def async_switch_home_schedule(self, home_id: str, schedule_id: str) -> Any:
        """Switch the schedule for a give home ID."""
        if not self.is_valid_schedule(home_id, schedule_id):
            raise NoSchedule("%s is not a valid schedule id" % schedule_id)

        post_params = {"home_id": home_id, "schedule_id": schedule_id}
        resp = await self.auth.async_post_request(
            url=_SWITCHHOMESCHEDULE_REQ,
            params=post_params,
        )
        LOG.debug("Response: %s", resp)


class AbstractHomeStatus(ABC):
    """Abstract class of the Netatmo home status."""

    raw_data: Dict = defaultdict(dict)
    rooms: Dict = defaultdict(dict)
    thermostats: Dict = defaultdict(dict)
    valves: Dict = defaultdict(dict)
    relays: Dict = defaultdict(dict)

    def process(self) -> None:
        """Process data from API."""
        for room in self.raw_data.get("rooms", []):
            self.rooms[room["id"]] = room

        for module in self.raw_data.get("modules", []):
            if module["type"] == "NATherm1":
                self.thermostats[module["id"]] = module

            elif module["type"] == "NRV":
                self.valves[module["id"]] = module

            elif module["type"] == "NAPlug":
                self.relays[module["id"]] = module

    def get_room(self, room_id: str) -> Dict:
        """Return room data for a given room id."""
        for key, value in self.rooms.items():
            if value["id"] == room_id:
                return self.rooms[key]

        raise InvalidRoom("No room with ID %s" % room_id)

    def get_thermostat(self, room_id: str) -> Dict:
        """Return thermostat data for a given room id."""
        for key, value in self.thermostats.items():
            if value["id"] == room_id:
                return self.thermostats[key]

        raise InvalidRoom("No room with ID %s" % room_id)

    def get_relay(self, room_id: str) -> Dict:
        """Return relay data for a given room id."""
        for key, value in self.relays.items():
            if value["id"] == room_id:
                return self.relays[key]

        raise InvalidRoom("No room with ID %s" % room_id)

    def get_valve(self, room_id: str) -> Dict:
        """Return valve data for a given room id."""
        for key, value in self.valves.items():
            if value["id"] == room_id:
                return self.valves[key]

        raise InvalidRoom("No room with ID %s" % room_id)

    def set_point(self, room_id: str) -> Optional[float]:
        """Return the setpoint of a given room."""
        return self.get_room(room_id).get("therm_setpoint_temperature")

    def set_point_mode(self, room_id: str) -> Optional[str]:
        """Return the setpointmode of a given room."""
        return self.get_room(room_id).get("therm_setpoint_mode")

    def measured_temperature(self, room_id: str) -> Optional[float]:
        """Return the measured temperature of a given room."""
        return self.get_room(room_id).get("therm_measured_temperature")

    def boiler_status(self, module_id: str) -> Optional[bool]:
        """Return the status of the boiler status."""
        return self.get_thermostat(module_id).get("boiler_status")


class HomeStatus(AbstractHomeStatus):
    """Class of the Netatmo home status."""

    def __init__(self, auth: NetatmoOAuth2, home_id: str):
        """Initialize the Netatmo home status.

        Arguments:
            auth {NetatmoOAuth2} -- Authentication information with a valid access token
            home_id {str} -- ID for targeted home
        """
        self.auth = auth

        self.home_id = home_id

    def update(self) -> None:
        """Fetch and process data from API."""
        post_params = {"home_id": self.home_id}

        resp = self.auth.post_request(url=_GETHOMESTATUS_REQ, params=post_params)
        if (
            "errors" in resp
            or "body" not in resp
            or "home" not in resp["body"]
            or ("errors" in resp["body"] and "modules" not in resp["body"]["home"])
        ):
            LOG.error("Errors in response: %s", resp)
            raise NoDevice("No device found, errors in response")

        self.raw_data = resp["body"]["home"]

        self.process()

    def set_thermmode(
        self,
        mode: str,
        end_time: int = None,
        schedule_id: str = None,
    ) -> Optional[str]:
        """Set thermotat mode."""
        post_params = {"home_id": self.home_id, "mode": mode}
        if end_time is not None and mode in ("hg", "away"):
            post_params["endtime"] = str(end_time)

        if schedule_id is not None and mode == "schedule":
            post_params["schedule_id"] = schedule_id

        return self.auth.post_request(url=_SETTHERMMODE_REQ, params=post_params)

    def set_room_thermpoint(
        self,
        room_id: str,
        mode: str,
        temp: float = None,
        end_time: int = None,
    ) -> Optional[str]:
        """Set room themperature set point."""
        post_params = {"home_id": self.home_id, "room_id": room_id, "mode": mode}
        # Temp and endtime should only be send when mode=='manual', but netatmo api can
        # handle that even when mode == 'home' and these settings don't make sense
        if temp is not None:
            post_params["temp"] = str(temp)

        if end_time is not None:
            post_params["endtime"] = str(end_time)

        return self.auth.post_request(url=_SETROOMTHERMPOINT_REQ, params=post_params)


class AsyncHomeStatus(AbstractHomeStatus):
    """Class of the Netatmo home status."""

    def __init__(self, auth: AbstractAsyncAuth, home_id: str):
        """Initialize the Netatmo home status.

        Arguments:
            auth {AbstractAsyncAuth} -- Authentication information with a valid access token
            home_id {str} -- ID for targeted home
        """
        self.auth = auth

        self.home_id = home_id

    async def async_update(self) -> None:
        """Fetch and process data from API."""
        post_params = {"home_id": self.home_id}

        resp = await self.auth.async_post_request(
            url=_GETHOMESTATUS_REQ,
            params=post_params,
        )
        if (
            "errors" in resp
            or "body" not in resp
            or "home" not in resp["body"]
            or ("errors" in resp["body"] and "modules" not in resp["body"]["home"])
        ):
            LOG.error("Errors in response: %s", resp)
            raise NoDevice("No device found, errors in response")

        self.raw_data = resp["body"]["home"]

        self.process()

    async def async_set_thermmode(
        self,
        mode: str,
        end_time: int = None,
        schedule_id: str = None,
    ) -> Optional[str]:
        """Set thermotat mode."""
        post_params = {"home_id": self.home_id, "mode": mode}
        if end_time is not None and mode in ("hg", "away"):
            post_params["endtime"] = str(end_time)

        if schedule_id is not None and mode == "schedule":
            post_params["schedule_id"] = schedule_id

        return await self.auth.async_post_request(
            url=_SETTHERMMODE_REQ,
            params=post_params,
        )

    async def async_set_room_thermpoint(
        self,
        room_id: str,
        mode: str,
        temp: float = None,
        end_time: int = None,
    ) -> Optional[str]:
        """Set room themperature set point."""
        post_params = {"home_id": self.home_id, "room_id": room_id, "mode": mode}
        # Temp and endtime should only be send when mode=='manual', but netatmo api can
        # handle that even when mode == 'home' and these settings don't make sense
        if temp is not None:
            post_params["temp"] = str(temp)

        if end_time is not None:
            post_params["endtime"] = str(end_time)

        return await self.auth.async_post_request(
            url=_SETROOMTHERMPOINT_REQ,
            params=post_params,
        )



"""Support for Netatmo authentication."""
import logging
from abc import ABC, abstractmethod
from json import JSONDecodeError
from time import sleep
from typing import Any, Callable, Dict, Optional, Tuple, Union

import requests
from aiohttp import ClientError, ClientSession
from oauthlib.oauth2 import LegacyApplicationClient, TokenExpiredError
from requests_oauthlib import OAuth2Session

from pyatmo.exceptions import ApiError
from pyatmo.helpers import _BASE_URL, ERRORS

LOG = logging.getLogger(__name__)

# Common definitions
AUTH_REQ_ENDPOINT = "oauth2/token"
AUTH_REQ = _BASE_URL + AUTH_REQ_ENDPOINT
AUTH_URL_ENDPOINT = "oauth2/authorize"
AUTH_URL = _BASE_URL + AUTH_URL_ENDPOINT
WEBHOOK_URL_ADD = _BASE_URL + "api/addwebhook"
WEBHOOK_URL_DROP = _BASE_URL + "api/dropwebhook"

AUTHORIZATION_HEADER = "Authorization"


# Possible scops
ALL_SCOPES = [
    "read_station",
    "read_camera",
    "access_camera",
    "write_camera",
    "read_presence",
    "access_presence",
    "write_presence",
    "read_homecoach",
    "read_smokedetector",
    "read_thermostat",
    "write_thermostat",
]


class NetatmoOAuth2:
    """
    Handle authentication with OAuth2
    """

    def __init__(
        self,
        client_id: str = None,
        client_secret: str = None,
        redirect_uri: Optional[str] = None,
        token: Optional[Dict[str, str]] = None,
        token_updater: Optional[Callable[[str], None]] = None,
        scope: Optional[str] = "read_station",
    ) -> None:
        """Initialize self.

        Keyword Arguments:
            client_id {str} -- Application client ID delivered by Netatmo on dev.netatmo.com (default: {None})
            client_secret {str} -- Application client secret delivered by Netatmo on dev.netatmo.com (default: {None})
            redirect_uri {Optional[str]} -- Redirect URI where to the authorization server will redirect with an authorization code (default: {None})
            token {Optional[Dict[str, str]]} -- Authorization token (default: {None})
            token_updater {Optional[Callable[[str], None]]} -- Callback when the token is updated (default: {None})
            scope {Optional[str]} -- List of scopes (default: {"read_station"})
                read_station: to retrieve weather station data (Getstationsdata, Getmeasure)
                read_camera: to retrieve Welcome data (Gethomedata, Getcamerapicture)
                access_camera: to access the camera, the videos and the live stream
                write_camera: to set home/away status of persons (Setpersonsaway, Setpersonshome)
                read_thermostat: to retrieve thermostat data (Getmeasure, Getthermostatsdata)
                write_thermostat: to set up the thermostat (Syncschedule, Setthermpoint)
                read_presence: to retrieve Presence data (Gethomedata, Getcamerapicture)
                access_presence: to access the live stream, any video stored on the SD card and to retrieve Presence's lightflood status
                read_homecoach: to retrieve Home Coache data (Gethomecoachsdata)
                read_smokedetector: to retrieve the smoke detector status (Gethomedata)
                Several values can be used at the same time, ie: 'read_station read_camera'
        """
        self.client_id = client_id
        self.client_secret = client_secret
        self.redirect_uri = redirect_uri
        self.token_updater = token_updater

        if token:
            self.scope = " ".join(token["scope"])

        else:
            self.scope = " ".join(ALL_SCOPES) if not scope else scope

        self.extra = {"client_id": self.client_id, "client_secret": self.client_secret}

        self._oauth = OAuth2Session(
            client_id=self.client_id,
            token=token,
            token_updater=self.token_updater,
            redirect_uri=self.redirect_uri,
            scope=self.scope,
        )

    def refresh_tokens(self) -> Dict[str, Union[str, int]]:
        """Refresh and return new tokens."""
        token = self._oauth.refresh_token(AUTH_REQ, **self.extra)

        if self.token_updater is not None:
            self.token_updater(token)

        return token

    def post_request(
        self,
        url: str,
        params: Optional[Dict] = None,
        timeout: int = 5,
    ) -> Any:
        """Wrapper for post requests."""
        resp = None
        req_args = {"data": params if params is not None else {}}

        if "json" in req_args["data"]:
            req_args["json"] = req_args["data"]["json"]
            req_args.pop("data")

        if "https://" not in url:
            try:
                resp = requests.post(url, data=params, timeout=timeout)
            except requests.exceptions.ChunkedEncodingError:
                LOG.debug("Encoding error when connecting to '%s'", url)
            except requests.exceptions.ConnectTimeout:
                LOG.debug("Connection to %s timed out", url)
            except requests.exceptions.ConnectionError:
                LOG.debug("Remote end closed connection without response (%s)", url)

        else:

            def query(url: str, params: Dict, timeout: int, retries: int) -> Any:
                if retries == 0:
                    LOG.error("Too many retries")
                    return

                try:
                    return self._oauth.post(url=url, timeout=timeout, **params)

                except (
                    TokenExpiredError,
                    requests.exceptions.ReadTimeout,
                    requests.exceptions.ConnectionError,
                ):
                    self._oauth.token = self.refresh_tokens()
                    # Sleep for 1 sec to prevent authentication related
                    # timeouts after a token refresh.
                    sleep(1)
                    return query(url, params, timeout * 2, retries - 1)

            resp = query(url, req_args, timeout, 3)

        if resp is None:
            LOG.debug("Resp is None - %s", resp)
            return None

        if not resp.ok:
            LOG.debug(
                "The Netatmo API returned %s (%s)",
                resp.content,
                resp.status_code,
            )
            try:
                raise ApiError(
                    f"{resp.status_code} - "
                    f"{ERRORS.get(resp.status_code, '')} - "
                    f"{resp.json()['error']['message']} "
                    f"({resp.json()['error']['code']}) "
                    f"when accessing '{url}'",
                )

            except JSONDecodeError as exc:
                raise ApiError(
                    f"{resp.status_code} - "
                    f"{ERRORS.get(resp.status_code, '')} - "
                    f"when accessing '{url}'",
                ) from exc

        try:
            if "application/json" in resp.headers.get("content-type", []):
                return resp.json()

            if resp.content not in [b"", b"None"]:
                return resp.content

        except (TypeError, AttributeError):
            LOG.debug("Invalid response %s", resp)

        return None

    def get_authorization_url(self, state: Optional[str] = None) -> Tuple[str, str]:
        return self._oauth.authorization_url(AUTH_URL, state)

    def request_token(
        self,
        authorization_response: Optional[str] = None,
        code: Optional[str] = None,
    ) -> Dict[str, str]:
        """
        Generic method for fetching a Netatmo access token.
        :param authorization_response: Authorization response URL, the callback
                                       URL of the request back to you.
        :param code: Authorization code
        :return: A token dict
        """
        return self._oauth.fetch_token(
            AUTH_REQ,
            authorization_response=authorization_response,
            code=code,
            client_secret=self.client_secret,
            include_client_id=True,
        )

    def addwebhook(self, webhook_url: str) -> None:
        post_params = {"url": webhook_url}
        resp = self.post_request(WEBHOOK_URL_ADD, post_params)
        LOG.debug("addwebhook: %s", resp)

    def dropwebhook(self) -> None:
        post_params = {"app_types": "app_security"}
        resp = self.post_request(WEBHOOK_URL_DROP, post_params)
        LOG.debug("dropwebhook: %s", resp)


class ClientAuth(NetatmoOAuth2):
    """
    Request authentication and keep access token available through token method. Renew it automatically if necessary
    Args:
        clientId (str): Application clientId delivered by Netatmo on dev.netatmo.com
        clientSecret (str): Application Secret key delivered by Netatmo on dev.netatmo.com
        username (str)
        password (str)
        scope (Optional[str]):
            read_station: to retrieve weather station data (Getstationsdata, Getmeasure)
            read_camera: to retrieve Welcome data (Gethomedata, Getcamerapicture)
            access_camera: to access the camera, the videos and the live stream
            write_camera: to set home/away status of persons (Setpersonsaway, Setpersonshome)
            read_thermostat: to retrieve thermostat data (Getmeasure, Getthermostatsdata)
            write_thermostat: to set up the thermostat (Syncschedule, Setthermpoint)
            read_presence: to retrieve Presence data (Gethomedata, Getcamerapicture)
            access_presence: to access the live stream, any video stored on the SD card and to retrieve Presence's lightflood status
            read_homecoach: to retrieve Home Coache data (Gethomecoachsdata)
            read_smokedetector: to retrieve the smoke detector status (Gethomedata)
            Several value can be used at the same time, ie: 'read_station read_camera'
    """

    def __init__(
        self,
        client_id: str,
        client_secret: str,
        username: str,
        password: str,
        scope="read_station",
    ):
        super().__init__(client_id=client_id, client_secret=client_secret, scope=scope)

        self._oauth = OAuth2Session(
            client=LegacyApplicationClient(client_id=self.client_id),
        )
        self._oauth.fetch_token(
            token_url=AUTH_REQ,
            username=username,
            password=password,
            client_id=self.client_id,
            client_secret=self.client_secret,
            scope=self.scope,
        )


class AbstractAsyncAuth(ABC):
    """Abstract class to make authenticated requests."""

    def __init__(self, websession: ClientSession):
        """Initialize the auth."""
        self.websession = websession

    @abstractmethod
    async def async_get_access_token(self) -> str:
        """Return a valid access token."""

    async def async_post_request(
        self,
        url: str,
        params: Optional[Dict] = None,
        timeout: int = 5,
    ) -> Any:
        """Wrapper for async post requests."""
        try:
            access_token = await self.async_get_access_token()
        except ClientError as err:
            raise ApiError(f"Access token failure: {err}") from err
        headers = {AUTHORIZATION_HEADER: f"Bearer {access_token}"}

        req_args = {"data": params if params is not None else {}}

        if "json" in req_args["data"]:
            req_args["json"] = req_args["data"]["json"]
            req_args.pop("data")

        async with self.websession.post(
            url,
            **req_args,
            headers=headers,
            timeout=timeout,
        ) as resp:
            resp_status = resp.status
            if resp.headers.get("content-type") == "image/jpeg":
                return await resp.read()

            resp_json = await resp.json()
            resp_content = await resp.read()

            if not resp.ok:
                LOG.debug("The Netatmo API returned %s (%s)", resp_content, resp_status)
                try:
                    raise ApiError(
                        f"{resp_status} - "
                        f"{ERRORS.get(resp_status, '')} - "
                        f"{resp_json['error']['message']} "
                        f"({resp_json['error']['code']}) "
                        f"when accessing '{url}'",
                    )

                except JSONDecodeError as exc:
                    raise ApiError(
                        f"{resp_status} - "
                        f"{ERRORS.get(resp_status, '')} - "
                        f"when accessing '{url}'",
                    ) from exc

            try:
                if "application/json" in resp.headers.get("content-type", []):
                    return resp_json

                if resp_content not in [b"", b"None"]:
                    return resp_content

            except (TypeError, AttributeError):
                LOG.debug("Invalid response %s", resp)

        return None

    async def async_addwebhook(self, webhook_url: str) -> None:
        """Register webhook."""
        resp = await self.async_post_request(WEBHOOK_URL_ADD, {"url": webhook_url})
        LOG.debug("addwebhook: %s", resp)

    async def async_dropwebhook(self) -> None:
        """Unregister webhook."""
        resp = await self.async_post_request(
            WEBHOOK_URL_DROP,
            {"app_types": "app_security"},
        )
        LOG.debug("dropwebhook: %s", resp)



from .auth import AbstractAsyncAuth, NetatmoOAuth2
from .helpers import _BASE_URL
from .weather_station import AsyncWeatherStationData, WeatherStationData

_GETHOMECOACHDATA_REQ = _BASE_URL + "api/gethomecoachsdata"


class HomeCoachData(WeatherStationData):
    """
    Class of Netatmo Home Couch devices (stations and modules)
    """

    def __init__(self, auth: NetatmoOAuth2) -> None:
        """Initialize self.

        Arguments:
            auth {NetatmoOAuth2} -- Authentication information with a valid access token
        """
        super().__init__(auth, url_req=_GETHOMECOACHDATA_REQ)


class AsyncHomeCoachData(AsyncWeatherStationData):
    """
    Class of Netatmo Home Couch devices (stations and modules)
    """

    def __init__(self, auth: AbstractAsyncAuth) -> None:
        """Initialize self.

        Arguments:
            auth {AbstractAsyncAuth} -- Authentication information with a valid access token
        """
        super().__init__(auth, url_req=_GETHOMECOACHDATA_REQ)



import os
import sys

from pyatmo.auth import ALL_SCOPES, ClientAuth
from pyatmo.camera import CameraData
from pyatmo.exceptions import NoDevice
from pyatmo.public_data import PublicData
from pyatmo.thermostat import HomeData
from pyatmo.weather_station import WeatherStationData

LON_NE = 6.221652
LAT_NE = 46.610870
LON_SW = 6.217828
LAT_SW = 46.596485


def main():
    try:
        if (
            os.environ["CLIENT_ID"]
            and os.environ["CLIENT_SECRET"]
            and os.environ["USERNAME"]
            and os.environ["PASSWORD"]
        ):
            client_id = os.environ["CLIENT_ID"]
            client_secret = os.environ["CLIENT_SECRET"]
            username = os.environ["USERNAME"]
            password = os.environ["PASSWORD"]
    except KeyError:
        sys.stderr.write(
            "No credentials passed to pyatmo.py (client_id, client_secret, "
            "username, password)\n",
        )
        sys.exit(1)

    auth = ClientAuth(
        client_id=client_id,
        client_secret=client_secret,
        username=username,
        password=password,
        scope=" ".join(ALL_SCOPES),
    )

    try:
        WeatherStationData(auth)
    except NoDevice:
        if sys.stdout.isatty():
            print("pyatmo.py : warning, no weather station available for testing")

    try:
        CameraData(auth)
    except NoDevice:
        if sys.stdout.isatty():
            print("pyatmo.py : warning, no camera available for testing")

    try:
        HomeData(auth)
    except NoDevice:
        if sys.stdout.isatty():
            print("pyatmo.py : warning, no thermostat available for testing")

    PublicData(auth, LAT_NE, LON_NE, LAT_SW, LON_SW)

    # If we reach this line, all is OK

    # If launched interactively, display OK message
    if sys.stdout.isatty():
        print("pyatmo: OK")

    sys.exit(0)


if __name__ == "__main__":
    main()



import logging
import time
from calendar import timegm
from datetime import datetime
from typing import Dict, Tuple

LOG: logging.Logger = logging.getLogger(__name__)

_BASE_URL: str = "https://api.netatmo.com/"

ERRORS: Dict[int, str] = {
    400: "Bad request",
    401: "Unauthorized",
    403: "Forbidden",
    404: "Not found",
    406: "Not Acceptable",
    500: "Internal Server Error",
    502: "Bad Gateway",
    503: "Service Unavailable",
}


def to_time_string(value: str) -> str:
    return datetime.utcfromtimestamp(int(value)).isoformat(sep="_")


def to_epoch(value: str) -> int:
    return timegm(time.strptime(value + "GMT", "%Y-%m-%d_%H:%M:%S%Z"))


def today_stamps() -> Tuple[int, int]:
    today: int = timegm(time.strptime(time.strftime("%Y-%m-%d") + "GMT", "%Y-%m-%d%Z"))
    return today, today + 3600 * 24


def fix_id(raw_data: Dict) -> Dict:
    if raw_data:
        for station in raw_data:
            station["_id"] = station["_id"].replace(" ", "")
            for module in station.get("modules", {}):
                module["_id"] = module["_id"].replace(" ", "")
    return raw_data



"""Support for Netatmo security devices (cameras, smoke detectors, sirens, window sensors, events and persons)."""
import imghdr
import time
from abc import ABC
from collections import defaultdict
from typing import Any, Dict, List, Optional, Tuple

from requests.exceptions import ReadTimeout

from .auth import AbstractAsyncAuth, NetatmoOAuth2
from .exceptions import ApiError, NoDevice
from .helpers import _BASE_URL, LOG

_GETHOMEDATA_REQ = _BASE_URL + "api/gethomedata"
_GETCAMERAPICTURE_REQ = _BASE_URL + "api/getcamerapicture"
_GETEVENTSUNTIL_REQ = _BASE_URL + "api/geteventsuntil"
_SETPERSONSAWAY_REQ = _BASE_URL + "api/setpersonsaway"
_SETPERSONSHOME_REQ = _BASE_URL + "api/setpersonshome"
_SETSTATE_REQ = _BASE_URL + "api/setstate"


class AbstractCameraData(ABC):
    """Abstract class of Netatmo camera data."""

    raw_data: Dict = defaultdict(dict)
    homes: Dict = defaultdict(dict)
    persons: Dict = {}
    events: Dict = defaultdict(dict)
    outdoor_events: Dict = defaultdict(dict)
    cameras: Dict = defaultdict(dict)
    smoke_detectors: Dict = defaultdict(dict)
    modules: Dict = {}
    last_event: Dict = {}
    outdoor_last_event: Dict = {}
    types: Dict = defaultdict(dict)

    def process(self) -> None:
        """Process data from API."""
        self.homes = {d["id"]: d for d in self.raw_data}

        for item in self.raw_data:
            home_id: str = item.get("id", "")

            if not item.get("name"):
                self.homes[home_id]["name"] = "Unknown"

            for person in item.get("persons", []):
                self.persons[person["id"]] = person

            for event in item.get("events", []):
                self._store_events(event)

            for camera in item.get("cameras", []):
                self.cameras[home_id][camera["id"]] = camera
                self.types[home_id][camera["type"]] = camera

                self.cameras[home_id][camera["id"]]["home_id"] = home_id
                if camera["type"] == "NACamera":
                    for module in camera.get("modules", []):
                        self.modules[module["id"]] = module
                        self.modules[module["id"]]["cam_id"] = camera["id"]

            for smoke in item.get("smokedetectors", []):
                self.smoke_detectors[home_id][smoke["id"]] = smoke
                self.types[home_id][smoke["type"]] = smoke

    def _store_events(self, event):
        if event["type"] == "outdoor":
            self.outdoor_events[event["camera_id"]][event["time"]] = event

        else:
            self.events[event["camera_id"]][event["time"]] = event

    def _store_last_event(self):
        for camera in self.events:
            self.last_event[camera] = self.events[camera][
                sorted(self.events[camera])[-1]
            ]

        for camera in self.outdoor_events:
            self.outdoor_last_event[camera] = self.outdoor_events[camera][
                sorted(self.outdoor_events[camera])[-1]
            ]

    def get_camera(self, camera_id: str) -> Dict[str, str]:
        """Get camera data."""
        for home_id in self.cameras:
            if camera_id in self.cameras[home_id]:
                return self.cameras[home_id][camera_id]

        return {}

    def get_module(self, module_id: str) -> Optional[dict]:
        """Get module data."""
        return None if module_id not in self.modules else self.modules[module_id]

    def get_smokedetector(self, smoke_id: str) -> Optional[dict]:
        """Get smoke detector."""
        for home_id in self.smoke_detectors:
            if smoke_id in self.smoke_detectors[home_id]:
                return self.smoke_detectors[home_id][smoke_id]

        return None

    def camera_urls(self, camera_id: str) -> Tuple[Optional[str], Optional[str]]:
        """
        Return the vpn_url and the local_url (if available) of a given camera
        in order to access its live feed.
        """
        camera_data = self.get_camera(camera_id)
        return camera_data.get("vpn_url", None), camera_data.get("local_url", None)

    def get_light_state(self, camera_id: str) -> Optional[str]:
        """Return the current mode of the floodlight of a presence camera."""
        camera_data = self.get_camera(camera_id)
        if camera_data is None:
            raise ValueError("Invalid Camera ID")

        return camera_data.get("light_mode_status")

    def persons_at_home(self, home_id: str = None) -> List:
        """Return a list of known persons who are currently at home."""
        home_data = self.homes.get(home_id, {})
        return [
            person["pseudo"]
            for person in home_data.get("persons")
            if "pseudo" in person and not person["out_of_sight"]
        ]

    def get_person_id(self, name: str) -> Optional[str]:
        """Retrieve the ID of a person."""
        for pid, data in self.persons.items():
            if name == data.get("pseudo"):
                return pid

        return None

    def build_event_id(self, event_id: Optional[str], device_type: Optional[str]):
        """Build event id."""

        def get_event_id(data: Dict):
            events = {e["time"]: e for e in data.values()}
            return min(events.items())[1].get("id")

        if not event_id:
            # If no event is provided we need to retrieve the oldest of
            # the last event seen by each camera
            if device_type == "NACamera":
                # for the Welcome camera
                event_id = get_event_id(self.last_event)

            elif device_type in {"NOC", "NSD"}:
                # for the Presence camera and for the smoke detector
                event_id = get_event_id(self.outdoor_last_event)

        return event_id

    def person_seen_by_camera(
        self,
        name: str,
        camera_id: str,
        exclude: int = 0,
    ) -> bool:
        """Evaluate if a specific person has been seen."""
        # Check in the last event is someone known has been seen
        def _person_in_event(curr_event, person_name):
            if curr_event["type"] == "person":
                person_id = curr_event["person_id"]

                if self.persons[person_id].get("pseudo") == person_name:
                    return True
            return None

        if exclude:
            limit = time.time() - exclude
            array_time_event = sorted(self.events[camera_id], reverse=True)

            for time_ev in array_time_event:
                if time_ev < limit:
                    return False

                current_event = self.events[camera_id][time_ev]
                if _person_in_event(current_event, name) is True:
                    return True

            return False

        current_event = self.last_event[camera_id]
        if _person_in_event(current_event, name) is True:
            return True

        return False

    def _known_persons(self) -> Dict[str, Dict]:
        """Return all known persons."""
        return {pid: p for pid, p in self.persons.items() if "pseudo" in p}

    def known_persons(self) -> Dict[str, str]:
        """Return a dictionary of known person names."""
        return {pid: p["pseudo"] for pid, p in self._known_persons().items()}

    def known_persons_names(self) -> List[str]:
        """Return a list of known person names."""
        return [person["pseudo"] for person in self._known_persons().values()]

    def someone_known_seen(self, camera_id: str, exclude: int = 0) -> bool:
        """Evaluate if someone known has been seen."""
        if camera_id not in self.events:
            raise NoDevice

        if exclude:
            limit = time.time() - exclude
            array_time_event = sorted(self.events[camera_id], reverse=True)

            for time_ev in array_time_event:
                if time_ev < limit:
                    return False

                curr_event = self.events[camera_id][time_ev]
                if curr_event["type"] == "person":
                    if curr_event["person_id"] in self._known_persons():
                        return True

        # Check in the last event if someone known has been seen
        else:
            curr_event = self.last_event[camera_id]
            if curr_event["type"] == "person":
                if curr_event["person_id"] in self._known_persons():
                    return True

        return False

    def someone_unknown_seen(self, camera_id: str, exclude: int = 0) -> bool:
        """Evaluate if someone known has been seen."""
        if camera_id not in self.events:
            raise NoDevice

        if exclude:
            limit = time.time() - exclude
            array_time_event = sorted(self.events[camera_id], reverse=True)

            for time_ev in array_time_event:
                if time_ev < limit:
                    return False

                curr_event = self.events[camera_id][time_ev]
                if curr_event["type"] == "person":
                    if curr_event["person_id"] not in self._known_persons():
                        return True

        # Check in the last event is noone known has been seen
        else:
            curr_event = self.last_event[camera_id]
            if curr_event["type"] == "person":
                if curr_event["person_id"] not in self._known_persons():
                    return True

        return False

    def motion_detected(self, camera_id: str, exclude: int = 0) -> bool:
        """Evaluate if movement has been detected."""
        if camera_id not in self.events:
            raise NoDevice

        if exclude:
            limit = time.time() - exclude
            array_time_event = sorted(self.events[camera_id], reverse=True)

            for time_ev in array_time_event:
                if time_ev < limit:
                    return False

                if self.events[camera_id][time_ev]["type"] == "movement":
                    return True

        elif self.last_event[camera_id]["type"] == "movement":
            return True

        return False

    def outdoor_motion_detected(self, camera_id: str, offset: int = 0) -> bool:
        """Evaluate if outdoor movement has been detected."""
        if camera_id not in self.last_event:
            return False

        last_event = self.last_event[camera_id]
        return (
            last_event["type"] == "movement"
            and last_event["video_status"] == "recording"
            and last_event["time"] + offset > int(time.time())
        )

    def _object_detected(self, object_name: str, camera_id: str, offset: int) -> bool:
        """Evaluate if a human has been detected."""
        if self.outdoor_last_event[camera_id]["video_status"] == "recording":
            for event in self.outdoor_last_event[camera_id]["event_list"]:
                if event["type"] == object_name and (
                    event["time"] + offset > int(time.time())
                ):
                    return True

        return False

    def human_detected(self, camera_id: str, offset: int = 0) -> bool:
        """Evaluate if a human has been detected."""
        return self._object_detected("human", camera_id, offset)

    def animal_detected(self, camera_id: str, offset: int = 0) -> bool:
        """Evaluate if an animal has been detected."""
        return self._object_detected("animal", camera_id, offset)

    def car_detected(self, camera_id: str, offset: int = 0) -> bool:
        """Evaluate if a car has been detected."""
        return self._object_detected("vehicle", camera_id, offset)

    def module_motion_detected(
        self,
        module_id: str,
        camera_id: str,
        exclude: int = 0,
    ) -> bool:
        """Evaluate if movement has been detected."""
        if exclude:
            limit = time.time() - exclude
            array_time_event = sorted(self.events.get(camera_id, []), reverse=True)

            for time_ev in array_time_event:
                if time_ev < limit:
                    return False

                curr_event = self.events[camera_id][time_ev]
                if (
                    curr_event["type"] in {"tag_big_move", "tag_small_move"}
                    and curr_event["module_id"] == module_id
                ):
                    return True

        else:
            if camera_id not in self.last_event:
                return False

            curr_event = self.last_event[camera_id]
            if (
                curr_event["type"] in {"tag_big_move", "tag_small_move"}
                and curr_event["module_id"] == module_id
            ):
                return True

        return False

    def module_opened(self, module_id: str, camera_id: str, exclude: int = 0) -> bool:
        """Evaluate if module status is open."""
        if exclude:
            limit = time.time() - exclude
            array_time_event = sorted(self.events.get(camera_id, []), reverse=True)

            for time_ev in array_time_event:
                if time_ev < limit:
                    return False

                curr_event = self.events[camera_id][time_ev]
                if (
                    curr_event["type"] == "tag_open"
                    and curr_event["module_id"] == module_id
                ):
                    return True

        else:
            if camera_id not in self.last_event:
                return False

            curr_event = self.last_event[camera_id]
            if (
                curr_event["type"] == "tag_open"
                and curr_event["module_id"] == module_id
            ):
                return True

        return False

    def build_state_params(
        self,
        camera_id: str,
        home_id: Optional[str],
        floodlight: Optional[str],
        monitoring: Optional[str],
    ):
        """."""
        if home_id is None:
            home_id = self.get_camera(camera_id)["home_id"]

        module = {"id": camera_id}

        if floodlight:
            param, val = "floodlight", floodlight.lower()
            if val not in {"on", "off", "auto"}:
                LOG.error("Invalid value for floodlight")
            else:
                module[param] = val

        if monitoring:
            param, val = "monitoring", monitoring.lower()
            if val not in {"on", "off"}:
                LOG.error("Invalid value for monitoring")
            else:
                module[param] = val

        return {"id": home_id, "modules": [module]}


class CameraData(AbstractCameraData):
    """Class of Netatmo camera data."""

    def __init__(self, auth: NetatmoOAuth2) -> None:
        """Initialize the Netatmo camera data.

        Arguments:
            auth {NetatmoOAuth2} -- Authentication information with a valid access token
        """
        self.auth = auth

    def update(self, events: int = 30) -> None:
        """Fetch and process data from API."""
        resp = self.auth.post_request(url=_GETHOMEDATA_REQ, params={"size": events})
        if resp is None or "body" not in resp:
            raise NoDevice("No device data returned by Netatmo server")

        self.raw_data = resp["body"].get("homes")
        if not self.raw_data:
            raise NoDevice("No device data available")

        self.process()
        self._update_all_camera_urls()
        self._store_last_event()

    def _update_all_camera_urls(self) -> None:
        """Update all camera urls."""
        for home_id in self.homes:
            for camera_id in self.cameras[home_id]:
                self.update_camera_urls(camera_id)

    def update_camera_urls(self, camera_id: str) -> None:
        """Update and validate the camera urls."""
        camera_data = self.get_camera(camera_id)
        home_id = camera_data["home_id"]

        if not camera_data or camera_data.get("status") == "disconnected":
            self.cameras[home_id][camera_id]["local_url"] = None
            self.cameras[home_id][camera_id]["vpn_url"] = None
            return

        vpn_url = camera_data.get("vpn_url")
        if vpn_url and camera_data.get("is_local"):

            def check_url(url: str) -> Optional[str]:
                try:
                    resp = self.auth.post_request(url=f"{url}/command/ping")
                except ReadTimeout:
                    LOG.debug("Timeout validation of camera url %s", url)
                    return None
                except ApiError:
                    LOG.debug("Api error for camera url %s", url)
                    return None

                return resp.get("local_url") if resp else None

            temp_local_url = check_url(vpn_url)
            if temp_local_url:
                self.cameras[home_id][camera_id]["local_url"] = check_url(
                    temp_local_url,
                )

    def set_state(
        self,
        camera_id: str,
        home_id: str = None,
        floodlight: str = None,
        monitoring: str = None,
    ) -> bool:
        """Turn camera (light) on/off.

        Arguments:
            camera_id {str} -- ID of a camera
            home_id {str} -- ID of a home
            floodlight {str} -- Mode for floodlight (on/off/auto)
            monitoring {str} -- Mode for monitoring (on/off)

        Returns:
            Boolean -- Success of the request
        """
        post_params = {
            "json": {
                "home": self.build_state_params(
                    camera_id,
                    home_id,
                    floodlight,
                    monitoring,
                ),
            },
        }

        try:
            resp = self.auth.post_request(url=_SETSTATE_REQ, params=post_params)
        except ApiError as err_msg:
            LOG.error("%s", err_msg)
            return False

        if "error" in resp:
            LOG.debug("%s", resp)
            return False

        LOG.debug("%s", resp)
        return True

    def set_persons_home(self, person_ids: List[str], home_id: str):
        """Mark persons as home."""
        post_params = {"home_id": home_id, "person_ids[]": person_ids}
        return self.auth.post_request(url=_SETPERSONSHOME_REQ, params=post_params)

    def set_persons_away(self, person_id: str, home_id: str):
        """Mark a person as away or set the whole home to being empty."""
        post_params = {"home_id": home_id, "person_id": person_id}
        return self.auth.post_request(url=_SETPERSONSAWAY_REQ, params=post_params)

    def get_camera_picture(
        self,
        image_id: str,
        key: str,
    ) -> Tuple[bytes, Optional[str]]:
        """Download a specific image (of an event or user face) from the camera."""
        post_params = {"image_id": image_id, "key": key}
        resp = self.auth.post_request(url=_GETCAMERAPICTURE_REQ, params=post_params)
        image_type = imghdr.what("NONE.FILE", resp)
        return resp, image_type

    def get_profile_image(self, name: str) -> Tuple[Optional[bytes], Optional[str]]:
        """Retrieve the face of a given person."""
        for person in self.persons:
            if name == self.persons[person].get("pseudo"):
                image_id = self.persons[person]["face"]["id"]
                key = self.persons[person]["face"]["key"]
                return self.get_camera_picture(image_id, key)

        return None, None

    def update_events(
        self,
        home_id: str,
        event_id: str = None,
        device_type: str = None,
    ) -> None:
        """Update the list of events."""
        if not (event_id or device_type):
            raise ApiError

        post_params = {
            "home_id": home_id,
            "event_id": self.build_event_id(event_id, device_type),
        }

        event_list: List = []
        resp: Optional[Dict[str, Any]] = None
        try:
            resp = self.auth.post_request(url=_GETEVENTSUNTIL_REQ, params=post_params)
            if resp is not None:
                event_list = resp["body"]["events_list"]
        except ApiError:
            pass
        except KeyError:
            if resp is not None:
                LOG.debug("event_list response: %s", resp)
                LOG.debug("event_list body: %s", dict(resp)["body"])
            else:
                LOG.debug("No resp received")

        for event in event_list:
            self._store_events(event)

        self._store_last_event()


class AsyncCameraData(AbstractCameraData):
    """Class of Netatmo camera data."""

    def __init__(self, auth: AbstractAsyncAuth) -> None:
        """Initialize the Netatmo camera data.

        Arguments:
            auth {AbstractAsyncAuth} -- Authentication information with a valid access token
        """
        self.auth = auth

    async def async_update(self, events: int = 30) -> None:
        """Fetch and process data from API."""
        resp = await self.auth.async_post_request(
            url=_GETHOMEDATA_REQ,
            params={"size": events},
        )
        if resp is None or "body" not in resp:
            raise NoDevice("No device data returned by Netatmo server")

        self.raw_data = resp["body"].get("homes")
        if not self.raw_data:
            raise NoDevice("No device data available")

        self.process()
        await self._async_update_all_camera_urls()
        self._store_last_event()

    async def _async_update_all_camera_urls(self) -> None:
        """Update all camera urls."""
        for home_id in self.homes:
            for camera_id in self.cameras[home_id]:
                await self.async_update_camera_urls(camera_id)

    async def async_set_state(
        self,
        camera_id: str,
        home_id: str = None,
        floodlight: str = None,
        monitoring: str = None,
    ) -> bool:
        """Turn camera (light) on/off.

        Arguments:
            camera_id {str} -- ID of a camera
            home_id {str} -- ID of a home
            floodlight {str} -- Mode for floodlight (on/off/auto)
            monitoring {str} -- Mode for monitoring (on/off)

        Returns:
            Boolean -- Success of the request
        """
        post_params = {
            "json": {
                "home": self.build_state_params(
                    camera_id,
                    home_id,
                    floodlight,
                    monitoring,
                ),
            },
        }

        try:
            resp = await self.auth.async_post_request(
                url=_SETSTATE_REQ,
                params=post_params,
            )
        except ApiError as err_msg:
            LOG.error("%s", err_msg)
            return False

        if "error" in resp:
            LOG.debug("%s", resp)
            return False

        LOG.debug("%s", resp)
        return True

    async def async_update_camera_urls(self, camera_id: str) -> None:
        """Update and validate the camera urls."""
        camera_data = self.get_camera(camera_id)
        home_id = camera_data["home_id"]

        if not camera_data or camera_data.get("status") == "disconnected":
            self.cameras[home_id][camera_id]["local_url"] = None
            self.cameras[home_id][camera_id]["vpn_url"] = None
            return

        vpn_url = camera_data.get("vpn_url")
        if vpn_url and camera_data.get("is_local"):

            async def async_check_url(url: str) -> Optional[str]:
                try:
                    resp = await self.auth.async_post_request(url=f"{url}/command/ping")
                except ReadTimeout:
                    LOG.debug("Timeout validation of camera url %s", url)
                    return None
                except ApiError:
                    LOG.debug("Api error for camera url %s", url)
                    return None

                return resp.get("local_url") if resp else None

            temp_local_url = await async_check_url(vpn_url)
            if temp_local_url:
                self.cameras[home_id][camera_id]["local_url"] = await async_check_url(
                    temp_local_url,
                )

    async def async_set_persons_home(self, person_ids: List[str], home_id: str):
        """Mark persons as home."""
        post_params = {"home_id": home_id, "person_ids[]": person_ids}
        return await self.auth.async_post_request(
            url=_SETPERSONSHOME_REQ,
            params=post_params,
        )

    async def async_set_persons_away(self, person_id: str, home_id: str):
        """Mark a person as away or set the whole home to being empty."""
        post_params = {"home_id": home_id, "person_id": person_id}
        return await self.auth.async_post_request(
            url=_SETPERSONSAWAY_REQ,
            params=post_params,
        )

    async def async_get_live_snapshot(self, camera_id: str):
        """Retrieve live snapshot from camera."""
        local, vpn = self.camera_urls(camera_id)
        if not local and not vpn:
            return None
        return await self.auth.async_post_request(
            url=f"{(local or vpn)}/live/snapshot_720.jpg",
            timeout=10,
        )

    async def async_get_camera_picture(
        self,
        image_id: str,
        key: str,
    ) -> Tuple[bytes, Optional[str]]:
        """Download a specific image (of an event or user face) from the camera."""
        post_params = {"image_id": image_id, "key": key}
        resp = await self.auth.async_post_request(
            url=_GETCAMERAPICTURE_REQ,
            params=post_params,
        )
        image_type = imghdr.what("NONE.FILE", resp)
        return resp, image_type

    async def async_get_profile_image(
        self,
        name: str,
    ) -> Tuple[Optional[bytes], Optional[str]]:
        """Retrieve the face of a given person."""
        for person in self.persons:
            if name == self.persons[person].get("pseudo"):
                image_id = self.persons[person]["face"]["id"]
                key = self.persons[person]["face"]["key"]
                return await self.async_get_camera_picture(image_id, key)

        return None, None



"""Expose submodules."""
from .auth import AbstractAsyncAuth, ClientAuth, NetatmoOAuth2
from .camera import AsyncCameraData, CameraData
from .exceptions import ApiError, InvalidHome, InvalidRoom, NoDevice, NoSchedule
from .home_coach import AsyncHomeCoachData, HomeCoachData
from .public_data import AsyncPublicData, PublicData
from .thermostat import AsyncHomeData, AsyncHomeStatus, HomeData, HomeStatus
from .weather_station import AsyncWeatherStationData, WeatherStationData

__all__ = [
    "AbstractAsyncAuth",
    "AsyncCameraData",
    "CameraData",
    "ClientAuth",
    "AsyncHomeCoachData",
    "HomeCoachData",
    "AsyncHomeData",
    "HomeData",
    "AsyncHomeStatus",
    "HomeStatus",
    "InvalidHome",
    "InvalidRoom",
    "ApiError",
    "NetatmoOAuth2",
    "NoDevice",
    "NoSchedule",
    "AsyncPublicData",
    "PublicData",
    "AsyncWeatherStationData",
    "WeatherStationData",
]



"""Support for Netatmo public weather data."""
import dataclasses
from abc import ABC
from collections import defaultdict
from typing import Any, Dict

from .auth import AbstractAsyncAuth, NetatmoOAuth2
from .exceptions import NoDevice
from .helpers import _BASE_URL, to_time_string

_GETPUBLIC_DATA = _BASE_URL + "api/getpublicdata"

_STATION_TEMPERATURE_TYPE = "temperature"
_STATION_PRESSURE_TYPE = "pressure"
_STATION_HUMIDITY_TYPE = "humidity"

_ACCESSORY_RAIN_LIVE_TYPE = "rain_live"
_ACCESSORY_RAIN_60MIN_TYPE = "rain_60min"
_ACCESSORY_RAIN_24H_TYPE = "rain_24h"
_ACCESSORY_RAIN_TIME_TYPE = "rain_timeutc"
_ACCESSORY_WIND_STRENGTH_TYPE = "wind_strength"
_ACCESSORY_WIND_ANGLE_TYPE = "wind_angle"
_ACCESSORY_WIND_TIME_TYPE = "wind_timeutc"
_ACCESSORY_GUST_STRENGTH_TYPE = "gust_strength"
_ACCESSORY_GUST_ANGLE_TYPE = "gust_angle"


@dataclasses.dataclass
class Location:
    """Class of Netatmo public weather location."""

    lat_ne: str
    lon_ne: str
    lat_sw: str
    lon_sw: str


class AbstractPublicData(ABC):
    """Class of Netatmo public weather data."""

    raw_data: dict = defaultdict(dict)
    status: str = ""
    time_exec: str = ""
    time_server: str = ""

    def process(self, resp: dict) -> None:
        """Process data from API."""
        self.status = resp["status"]
        self.time_exec = to_time_string(resp["time_exec"])
        self.time_server = to_time_string(resp["time_server"])

    def stations_in_area(self) -> int:
        return len(self.raw_data)

    def get_latest_rain(self) -> Dict:
        return self.get_accessory_data(_ACCESSORY_RAIN_LIVE_TYPE)

    def get_average_rain(self) -> float:
        return average(self.get_latest_rain())

    def get_60_min_rain(self) -> Dict:
        return self.get_accessory_data(_ACCESSORY_RAIN_60MIN_TYPE)

    def get_average_60_min_rain(self) -> float:
        return average(self.get_60_min_rain())

    def get_24_h_rain(self) -> Dict:
        return self.get_accessory_data(_ACCESSORY_RAIN_24H_TYPE)

    def get_average_24_h_rain(self) -> float:
        return average(self.get_24_h_rain())

    def get_latest_pressures(self) -> Dict:
        return self.get_latest_station_measures(_STATION_PRESSURE_TYPE)

    def get_average_pressure(self) -> float:
        return average(self.get_latest_pressures())

    def get_latest_temperatures(self) -> Dict:
        return self.get_latest_station_measures(_STATION_TEMPERATURE_TYPE)

    def get_average_temperature(self) -> float:
        return average(self.get_latest_temperatures())

    def get_latest_humidities(self) -> Dict:
        return self.get_latest_station_measures(_STATION_HUMIDITY_TYPE)

    def get_average_humidity(self) -> float:
        return average(self.get_latest_humidities())

    def get_latest_wind_strengths(self) -> Dict:
        return self.get_accessory_data(_ACCESSORY_WIND_STRENGTH_TYPE)

    def get_average_wind_strength(self) -> float:
        return average(self.get_latest_wind_strengths())

    def get_latest_wind_angles(self) -> Dict:
        return self.get_accessory_data(_ACCESSORY_WIND_ANGLE_TYPE)

    def get_latest_gust_strengths(self) -> Dict:
        return self.get_accessory_data(_ACCESSORY_GUST_STRENGTH_TYPE)

    def get_average_gust_strength(self) -> float:
        return average(self.get_latest_gust_strengths())

    def get_latest_gust_angles(self):
        return self.get_accessory_data(_ACCESSORY_GUST_ANGLE_TYPE)

    def get_locations(self) -> Dict:
        return {
            station["_id"]: station["place"]["location"] for station in self.raw_data
        }

    def get_time_for_rain_measures(self) -> Dict:
        return self.get_accessory_data(_ACCESSORY_RAIN_TIME_TYPE)

    def get_time_for_wind_measures(self) -> Dict:
        return self.get_accessory_data(_ACCESSORY_WIND_TIME_TYPE)

    def get_latest_station_measures(self, data_type) -> Dict:
        measures: Dict = {}
        for station in self.raw_data:
            for module in station["measures"].values():
                if (
                    "type" in module
                    and data_type in module["type"]
                    and "res" in module
                    and module["res"]
                ):
                    measure_index = module["type"].index(data_type)
                    latest_timestamp = sorted(module["res"], reverse=True)[0]
                    measures[station["_id"]] = module["res"][latest_timestamp][
                        measure_index
                    ]

        return measures

    def get_accessory_data(self, data_type: str) -> Dict[str, Any]:
        data: Dict = {}
        for station in self.raw_data:
            for module in station["measures"].values():
                if data_type in module:
                    data[station["_id"]] = module[data_type]

        return data


class PublicData(AbstractPublicData):
    """Class of Netatmo public weather data."""

    def __init__(
        self,
        auth: NetatmoOAuth2,
        lat_ne: str,
        lon_ne: str,
        lat_sw: str,
        lon_sw: str,
        required_data_type: str = None,
        filtering: bool = False,
    ) -> None:
        """Initialize self.

        Arguments:
            auth {NetatmoOAuth2} -- Authentication information with a valid access token
            LAT_NE {str} -- Latitude of the north east corner of the requested area. (-85 <= LAT_NE <= 85 and LAT_NE > LAT_SW)
            LON_NE {str} -- Longitude of the north east corner of the requested area. (-180 <= LON_NE <= 180 and LON_NE > LON_SW)
            LAT_SW {str} -- latitude of the south west corner of the requested area. (-85 <= LAT_SW <= 85)
            LON_SW {str} -- Longitude of the south west corner of the requested area. (-180 <= LON_SW <= 180)

        Keyword Arguments:
            required_data_type {str} -- comma-separated list from above _STATION or _ACCESSORY values (default: {None})
        """
        self.auth = auth
        self.required_data_type = required_data_type
        self.location = Location(lat_ne, lon_ne, lat_sw, lon_sw)
        self.filtering: bool = filtering

    def update(self) -> None:
        """Fetch and process data from API."""
        post_params: Dict = {
            **dataclasses.asdict(self.location),
            "filter": self.filtering,
        }

        if self.required_data_type:
            post_params["required_data"] = self.required_data_type

        resp = self.auth.post_request(url=_GETPUBLIC_DATA, params=post_params)
        try:
            self.raw_data = resp["body"]
        except (KeyError, TypeError) as exc:
            raise NoDevice("No public weather data returned by Netatmo server") from exc

        self.process(resp)


class AsyncPublicData(AbstractPublicData):
    """Class of Netatmo public weather data."""

    def __init__(
        self,
        auth: AbstractAsyncAuth,
        lat_ne: str,
        lon_ne: str,
        lat_sw: str,
        lon_sw: str,
        required_data_type: str = None,
        filtering: bool = False,
    ) -> None:
        """Initialize self.

        Arguments:
            auth {AbstractAsyncAuth} -- Authentication information with a valid access token
            LAT_NE {str} -- Latitude of the north east corner of the requested area. (-85 <= LAT_NE <= 85 and LAT_NE > LAT_SW)
            LON_NE {str} -- Longitude of the north east corner of the requested area. (-180 <= LON_NE <= 180 and LON_NE > LON_SW)
            LAT_SW {str} -- latitude of the south west corner of the requested area. (-85 <= LAT_SW <= 85)
            LON_SW {str} -- Longitude of the south west corner of the requested area. (-180 <= LON_SW <= 180)

        Keyword Arguments:
            required_data_type {str} -- comma-separated list from above _STATION or _ACCESSORY values (default: {None})
        """
        self.auth = auth
        self.required_data_type = required_data_type
        self.location = Location(lat_ne, lon_ne, lat_sw, lon_sw)
        self.filtering: bool = filtering

    async def async_update(self) -> None:
        """Fetch and process data from API."""
        post_params: Dict = {
            **dataclasses.asdict(self.location),
            "filter": self.filtering,
        }

        if self.required_data_type:
            post_params["required_data"] = self.required_data_type

        resp = await self.auth.async_post_request(
            url=_GETPUBLIC_DATA,
            params=post_params,
        )
        try:
            self.raw_data = resp["body"]
        except (KeyError, TypeError) as exc:
            raise NoDevice("No public weather data returned by Netatmo server") from exc

        self.process(resp)


def average(data: dict) -> float:
    return sum(data.values()) / len(data) if data else 0.0



"""Support for Netatmo weather station devices (stations and modules)."""
import logging
import time
from abc import ABC
from collections import defaultdict
from typing import Dict, List, Optional, Tuple

from .auth import AbstractAsyncAuth, NetatmoOAuth2
from .exceptions import NoDevice
from .helpers import _BASE_URL, fix_id, today_stamps

LOG = logging.getLogger(__name__)

_GETMEASURE_REQ = _BASE_URL + "api/getmeasure"
_GETSTATIONDATA_REQ = _BASE_URL + "api/getstationsdata"


class AbstractWeatherStationData(ABC):
    """Abstract class of Netatmo Weather Station devices."""

    raw_data: Dict = defaultdict(dict)
    stations: Dict = defaultdict(dict)
    modules: Dict = defaultdict(dict)

    def process(self) -> None:
        """Process data from API."""
        self.stations = {d["_id"]: d for d in self.raw_data}
        self.modules = {}

        for item in self.raw_data:
            # The station name is sometimes not contained in the backend data
            if "station_name" not in item:
                item["station_name"] = item.get("home_name", item["type"])

            if "modules" not in item:
                item["modules"] = [item]

            for module in item["modules"]:
                if "module_name" not in module and module["type"] == "NHC":
                    module["module_name"] = module["station_name"]

                self.modules[module["_id"]] = module
                self.modules[module["_id"]]["main_device"] = item["_id"]

    def get_module_names(self, station_id: str) -> List:
        """Return a list of all module names for a given station."""
        res = set()
        station_data = self.get_station(station_id)

        if not station_data:
            return []

        res.add(station_data.get("module_name", station_data.get("type")))
        for module in station_data["modules"]:
            # Add module name, use module type if no name is available
            res.add(module.get("module_name", module.get("type")))

        return list(res)

    def get_modules(self, station_id: str) -> Dict:
        """Return a dict of modules per given station."""
        station_data = self.get_station(station_id)

        if not station_data:
            return {}

        res = {}
        for station in [self.stations[station_data["_id"]]]:
            station_type = station.get("type")
            station_name = station.get("station_name", station_type)
            res[station["_id"]] = {
                "station_name": station_name,
                "module_name": station.get("module_name", station_type),
                "id": station["_id"],
            }

            for module in station["modules"]:
                res[module["_id"]] = {
                    "station_name": module.get("station_name", station_name),
                    "module_name": module.get("module_name", module.get("type")),
                    "id": module["_id"],
                }

        return res

    def get_station(self, station_id: str) -> Dict:
        """Return station by id."""
        return self.stations.get(station_id, {})

    def get_module(self, module_id: str) -> Dict:
        """Return module by id."""
        return self.modules.get(module_id, {})

    def get_monitored_conditions(self, module_id: str) -> List:
        """Return monitored conditions for given module."""
        module = self.get_module(module_id)
        if not module:
            module = self.get_station(module_id)

        if not module:
            return []

        conditions = []
        for condition in module.get("data_type", []):
            if condition == "Wind":
                # the Wind meter actually exposes the following conditions
                conditions.extend(
                    ["WindAngle", "WindStrength", "GustAngle", "GustStrength"],
                )

            elif condition == "Rain":
                conditions.extend(["Rain", "sum_rain_24", "sum_rain_1"])

            else:
                conditions.append(condition)

        if module["type"] in ["NAMain", "NHC"]:
            # the main module has wifi_status
            conditions.append("wifi_status")

        else:
            # assume all other modules have rf_status, battery_vp, and battery_percent
            conditions.extend(["rf_status", "battery_vp", "battery_percent"])

        if module["type"] in ["NAMain", "NAModule1", "NAModule4"]:
            conditions.extend(["temp_trend"])

        if module["type"] == "NAMain":
            conditions.extend(["pressure_trend"])

        if module["type"] in [
            "NAMain",
            "NAModule1",
            "NAModule2",
            "NAModule3",
            "NAModule4",
            "NHC",
        ]:
            conditions.append("reachable")

        return conditions

    def get_last_data(self, station_id: str, exclude: int = 0) -> Dict:
        """Return data for a given station and time frame."""
        key = "_id"

        last_data: Dict = {}
        station = self.get_station(station_id)

        if not station or "dashboard_data" not in station:
            LOG.debug("No dashboard data for station %s", station_id)
            return last_data

        # Define oldest acceptable sensor measure event
        limit = (time.time() - exclude) if exclude else 0

        data = station["dashboard_data"]
        if key in station and data["time_utc"] > limit:
            last_data[station[key]] = data.copy()
            last_data[station[key]]["When"] = last_data[station[key]].pop("time_utc")
            last_data[station[key]]["wifi_status"] = station.get("wifi_status")
            last_data[station[key]]["reachable"] = station.get("reachable")

        for module in station["modules"]:

            if "dashboard_data" not in module or key not in module:
                continue

            data = module["dashboard_data"]
            if "time_utc" in data and data["time_utc"] > limit:
                last_data[module[key]] = data.copy()
                last_data[module[key]]["When"] = last_data[module[key]].pop("time_utc")

                # For potential use, add battery and radio coverage information
                # to module data if present
                for i in (
                    "rf_status",
                    "battery_vp",
                    "battery_percent",
                    "reachable",
                    "wifi_status",
                ):
                    if i in module:
                        last_data[module[key]][i] = module[i]

        return last_data

    def check_not_updated(self, station_id: str, delay: int = 3600) -> List:
        """Check if a given station has not been updated."""
        res = self.get_last_data(station_id)
        return [
            key for key, value in res.items() if time.time() - value["When"] > delay
        ]

    def check_updated(self, station_id: str, delay: int = 3600) -> List:
        """Check if a given station has been updated."""
        res = self.get_last_data(station_id)
        return [
            key for key, value in res.items() if time.time() - value["When"] < delay
        ]


class WeatherStationData(AbstractWeatherStationData):
    """Class of Netatmo weather station devices."""

    def __init__(self, auth: NetatmoOAuth2, url_req: str = _GETSTATIONDATA_REQ) -> None:
        """Initialize the Netatmo weather station data.

        Arguments:
            auth {NetatmoOAuth2} -- Authentication information with a valid access token
            url_req {str} -- Optional request endpoint
        """
        self.auth = auth
        self.url_req = url_req

    def update(self):
        """Fetch data from API."""
        resp = self.auth.post_request(url=self.url_req)

        if resp is None or "body" not in resp:
            raise NoDevice("No weather station data returned by Netatmo server")

        try:
            self.raw_data = fix_id(resp["body"].get("devices"))
        except KeyError as exc:
            LOG.debug("No <body> in response %s", resp)
            raise NoDevice(
                "No weather station data returned by Netatmo server",
            ) from exc

        if not self.raw_data:
            raise NoDevice("No weather station available")

        self.process()

    def get_data(
        self,
        device_id: str,
        scale: str,
        module_type: str,
        module_id: str = None,
        date_begin: float = None,
        date_end: float = None,
        limit: int = None,
        optimize: bool = False,
        real_time: bool = False,
    ) -> Optional[Dict]:
        """Retrieve data from a device or module."""
        post_params = {"device_id": device_id}
        if module_id:
            post_params["module_id"] = module_id

        post_params["scale"] = scale
        post_params["type"] = module_type

        if date_begin:
            post_params["date_begin"] = f"{date_begin}"

        if date_end:
            post_params["date_end"] = f"{date_end}"

        if limit:
            post_params["limit"] = f"{limit}"

        post_params["optimize"] = "true" if optimize else "false"
        post_params["real_time"] = "true" if real_time else "false"

        return self.auth.post_request(url=_GETMEASURE_REQ, params=post_params)

    def get_min_max_t_h(
        self,
        station_id: str,
        module_id: str = None,
        frame: str = "last24",
    ) -> Optional[Tuple[float, float, float, float]]:
        """Return minimum and maximum temperature and humidity over the given timeframe.

        Arguments:
            station_id {str} -- Station ID

        Keyword Arguments:
            module_id {str} -- Module ID (default: {None})
            frame {str} -- Timeframe can be "last24" or "day" (default: {"last24"})

        Returns:
            (min_t {float}, max_t {float}, min_h {float}, max_h {float}) -- minimum and maximum for temperature and humidity
        """
        if frame == "last24":
            end = time.time()
            start = end - 24 * 3600  # 24 hours ago

        elif frame == "day":
            start, end = today_stamps()

        else:
            raise ValueError("'frame' value can only be 'last24' or 'day'")

        resp = self.get_data(
            device_id=station_id,
            module_id=module_id,
            scale="max",
            module_type="Temperature,Humidity",
            date_begin=start,
            date_end=end,
        )

        if resp:
            temperature = [temp[0] for temp in resp["body"].values()]
            humidity = [hum[1] for hum in resp["body"].values()]
            return min(temperature), max(temperature), min(humidity), max(humidity)

        return None


class AsyncWeatherStationData(AbstractWeatherStationData):
    """Class of Netatmo weather station devices."""

    def __init__(
        self,
        auth: AbstractAsyncAuth,
        url_req: str = _GETSTATIONDATA_REQ,
    ) -> None:
        """Initialize the Netatmo weather station data.

        Arguments:
            auth {AbstractAsyncAuth} -- Authentication information with a valid access token
            url_req {str} -- Optional request endpoint
        """
        self.auth = auth
        self.url_req = url_req

    async def async_update(self):
        """Fetch data from API."""
        resp = await self.auth.async_post_request(url=self.url_req)

        if resp is None or "body" not in resp:
            raise NoDevice("No weather station data returned by Netatmo server")

        try:
            self.raw_data = fix_id(resp["body"].get("devices"))
        except KeyError as exc:
            LOG.debug("No <body> in response %s", resp)
            raise NoDevice(
                "No weather station data returned by Netatmo server",
            ) from exc

        if not self.raw_data:
            raise NoDevice("No weather station available")

        self.process()
